#ifndef f3d_config_h
#define f3d_config_h

#include <string>

namespace f3d::detail
{
extern const std::string LibVersion;
extern const std::string LibVersionFull;
extern const std::string LibBuildSystem;
extern const std::string LibBuildDate;
extern const std::string LibCompiler;
}
#endif
