/**
 * @class   F3DPluginsTools
 * @brief   A namespace to load plugins
 *
 */
#ifndef F3DPluginsTools_h
#define F3DPluginsTools_h

#include <string>
#include <vector>

namespace F3DPluginsTools
{
void LoadPlugins(const std::vector<std::string>& plugins);
};

#endif
