LIBRARY "F3DShellExtension"

EXPORTS
    DllGetClassObject       PRIVATE
    DllCanUnloadNow         PRIVATE
    DllRegisterServer       PRIVATE
    DllUnregisterServer     PRIVATE
    DllNotifyShell          PRIVATE
