kbd {
    // TODO: use just the docs variables?
    --kbd-color-background: #e6e1e8;
    --kbd-color-border: #959396;
    --kbd-color-text: #27262b;

    // Background/text colors
    background-color: var(--kbd-color-background);
    color: var(--kbd-color-text);

    // Border/shadow
    border-radius: 0.25rem;
    background-clip: padding-box;
    border: 1px solid var(--kbd-color-border);
    box-shadow: 0 2px 0 1px var(--kbd-color-border);

    // Remove the text cursor
    cursor: default;

    // text style
    font-size: 0.75em;
    line-height: 1;
    min-width: 0.75rem;
    text-align: center;
    padding: 2px 5px;
    margin: 0 3px;
    position: relative;
    top: -1px;

    // Push effect on hover
    &:hover, li:hover & {
        box-shadow: 0 1px 0 0.5px var(--kbd-color-border);
        top: 1px;
    }

    // Transition is fun
    transition: all 0.1s;
}

div.install-buttons {
    text-align: center;
    display: block;
    line-height: 1;

    .btn {
        margin: .5em;
    }
}

img.cm {
    min-height: 10px;
}

div.popup-container {
    position: fixed;
    top: 25%;
    left: 50%;
    transform: translate(-50%, -50%);
    width: 500px;
    max-width: 75vw;
    padding: 1em 1.5em;
    box-shadow: 3px 2px 4px 2px rgba(0, 0, 0, 0.4);
    border-radius: .5em;
    
    .btn {
        display: block;
        margin-left: auto;
    }
}
